package mon.stat

import scala.reflect.runtime.universe._
import scala.tools.reflect.ToolBox

import scala.io.Source

import mon.util.YamlHelper

abstract class Status {
  val name : String
  // val id
  // val effectType
  val onStart : Option[Monster => Unit]
  // val onModifyStat
  // val onBeforeMovePriority : Int
  // val onBeforeMove
  // val onModifyMove
  // val onHit
  val onResidualOrder : Int
  val onResidual : Option[Monster => Unit]
  // val onSwitchIn
  
  override def toString = name
}

case class StatusToken(
    val name : String,
    val onStart : String,
    val onResidualOrder : Int,
    val onResidual : String
    ) {
  def instantiate() = {
    val self = this
    new Status{
      val name = self.name
      val onStart = Status.compileOnStart(self.onStart)
      val onResidualOrder = self.onResidualOrder
      val onResidual = Status.compileOnResidual(self.onResidual)
    }
  }
}

object Status {
  private var statuses = Map[String, Status]()
  val tokens = YamlHelper.extractSeq[StatusToken](Source.fromInputStream(Move.getClass.getResourceAsStream("data/statuses.yaml"))).map(t => (t.name, t)).toMap
  
  def apply(name : String) = {
    if (!statuses.contains(name)) {
      statuses = statuses.updated(name, tokens(name).instantiate())
    }
    statuses(name)
  }
  
  private val tb = runtimeMirror(getClass.getClassLoader).mkToolBox()
  
  def compileOnStart(code : String) : Option[(Monster /*, source, source effect */) => Unit] = {
    if (code == null) {
      None
    } else {
      val tree = tb.parse(
        s"""
           |import mon.stat.Monster
           |import mon.stat.Statistic._
           |def onStart(mon : Monster) = {
           |  $code
           |}
           |onStart _
          """.stripMargin)
      val f = tb.compile(tree)
      val wrapper = f()
      
      Some(wrapper.asInstanceOf[Monster => Unit])
    }
  }
  
  def compileOnResidual(code : String) : Option[Monster => Unit] = {
    if (code == null) {
      None
    } else {
      val tree = tb.parse(
        s"""
           |import mon.stat.Monster
           |import mon.stat.Statistic._
           |def onResidual(mon : Monster) = {
           |  $code
           |}
           |onResidual _
          """.stripMargin)
      val f = tb.compile(tree)
      val wrapper = f()
      
      Some(wrapper.asInstanceOf[Monster => Unit])
    }
  }
}